from flask_cognito_lib.config import Config
from flask_cognito_lib.services.cognito_svc import CognitoService
from flask_cognito_lib.services.token_svc import TokenService


def cognito_service_factory(cfg: Config):
    return CognitoService(cfg=cfg)


def token_service_factory(cfg: Config):
    return TokenService(cfg=cfg)



from typing import Any, List, Optional
from urllib.parse import quote

from flask import current_app

from .exceptions import ConfigurationError


def get(key: str, required: bool = False, default: Optional[Any] = None) -> Any:
    """Get a key from the current Flask application's configuration

    Parameters
    ----------
    key : str
        Name of key
    required : bool, optional
        Set whether it is required, by default False
    default : Optional[Any], optional
        A default value for `required=False` keys that are not already in the
        app config, by default None

    Returns
    -------
    Any
        The value of the configuration `key`

    Raises
    ------
    CognitoConfigurationError
        If key is required but no in the current app configuration
    """
    if key not in current_app.config and required:
        raise ConfigurationError("Missing required configuration parameter: ", key)

    if key not in current_app.config:
        return default

    return current_app.config[key]


class Config:
    """
    Helper class to hold the congfiguration
    """

    # Constants
    APP_EXTENSION_KEY = "cognito_auth_lib"
    CONTEXT_KEY_COGNITO_SERVICE = "aws_cognito_service"
    CONTEXT_KEY_TOKEN_SERVICE = "aws_jwt_service"
    COOKIE_NAME = "cognito_access_token"
    COOKIE_NAME_REFRESH = "cognito_refresh_token"

    @property
    def disabled(self) -> bool:
        """Return True if Cognito Authentication is disabled"""
        return get("AWS_COGNITO_DISABLED", required=False, default=False)

    @property
    def user_pool_id(self) -> str:
        """Return the Cognito user pool ID"""
        return get("AWS_COGNITO_USER_POOL_ID", required=True)

    @property
    def user_pool_client_id(self) -> str:
        """Return the Cognito user pool client ID"""
        return get("AWS_COGNITO_USER_POOL_CLIENT_ID", required=True)

    @property
    def user_pool_client_secret(self) -> str:
        """Return the Cognito user pool client secret"""
        return get("AWS_COGNITO_USER_POOL_CLIENT_SECRET", required=False)

    @property
    def redirect_url(self) -> str:
        """Return the Redirect URL (post-login)"""
        return get("AWS_COGNITO_REDIRECT_URL", required=True)

    @property
    def logout_redirect(self) -> str:
        """Return the Redirect URL (post-logout)"""
        return get("AWS_COGNITO_LOGOUT_URL", required=True)

    @property
    def domain(self) -> str:
        """Return the Cognito domain"""
        return get("AWS_COGNITO_DOMAIN", required=True)

    @property
    def region(self) -> str:
        """Return the AWS region"""
        return get("AWS_REGION", required=True)

    @property
    def max_cookie_age_seconds(self) -> int:
        """Return maximum age to keep an access token cookie for, in seconds"""
        return int(get("AWS_COGNITO_COOKIE_AGE_SECONDS", required=False, default=1800))

    @property
    def cognito_expiration_leeway(self) -> int:
        """Return the leeway (in seconds) for checking token expiration

        This is here largely for testing purposes. In production applications
        this should be set to zero.
        """
        return int(get("AWS_COGNITO_EXPIRATION_LEEWAY", required=False, default=0))

    @property
    def cognito_scopes(self) -> Optional[List[str]]:
        """
        Return the scopes to request from Cognito.
        If None, all supported scopes are returned
        """
        return get("AWS_COGNITO_SCOPES", required=False)

    @property
    def cookie_domain(self) -> str:
        """Return the domain used for the cookie.

        Used if you want to set a cross-domain cookie.
        For example, domain=".example.com" will set a cookie that is readable
        by the domain www.example.com, foo.example.com etc.

        If not set (default) then the cookie will only be readable by the
        domain that set it.
        """
        return get("AWS_COGNITO_COOKIE_DOMAIN", required=False)

    @property
    def cookie_samesite(self) -> str:
        """Return the property to set for "samesite" on the cookie

        The SameSite attribute lets servers specify whether/when cookies are
        sent with cross-site requests (where Site is defined by the registrable
        domain and the scheme: http or https). This provides some protection
        against cross-site request forgery attacks (CSRF).
        It takes three possible values: Strict, Lax, and None.
        """
        return get("AWS_COGNITO_COOKIE_SAMESITE", required=False)

    @property
    def refresh_flow_enabled(self) -> bool:
        """Return True if Cognito Refresh flow is enabled"""
        return get("AWS_COGNITO_REFRESH_FLOW_ENABLED", required=False, default=False)

    @property
    def refresh_cookie_encrypted(self) -> bool:
        """Return True if Cognito Refresh cookie should be encrypted"""
        return get("AWS_COGNITO_REFRESH_COOKIE_ENCRYPTED", required=False, default=True)

    @property
    def max_refresh_cookie_age_seconds(self) -> int:
        """Return maximum age to keep a refresh token cookie for, in seconds"""
        return int(
            get("AWS_COGNITO_REFRESH_COOKIE_AGE_SECONDS", required=False, default=86400)
        )

    @property
    def secret_key(self) -> bytes:
        """Return Flask secret key"""
        key = get("SECRET_KEY", required=True)
        if isinstance(key, str):
            return key.encode()
        return key

    @property
    def issuer(self) -> str:
        """Return the issuer"""
        return f"https://cognito-idp.{self.region}.amazonaws.com/{self.user_pool_id}"

    @property
    def jwk_endpoint(self) -> str:
        """Return the endpoint that holds the JWKs"""
        return f"{self.issuer}/.well-known/jwks.json"

    @property
    def token_endpoint(self) -> str:
        """Return the Cognito TOKEN endpoint URL"""
        return f"{self.domain}/oauth2/token"

    @property
    def authorize_endpoint(self) -> str:
        """Return the Cognito AUTHORIZE endpoint URL"""
        return f"{self.domain}/oauth2/authorize"

    @property
    def logout_endpoint(self) -> str:
        """Return the Cognito LOGOUT endpoint URL"""
        return (
            f"{self.domain}/logout"
            f"?client_id={self.user_pool_client_id}"
            f"&logout_uri={quote(self.logout_redirect)}"
        )

    @property
    def user_info_endpoint(self) -> str:
        """Return the Cognito USERINFO endpoint URL"""
        return f"{self.domain}/oauth2/userInfo"

    @property
    def revoke_endpoint(self) -> str:
        """Return the Cognito REVOKE endpoint URL"""
        return f"{self.domain}/oauth2/revoke"



from base64 import urlsafe_b64encode
from hashlib import sha256
from typing import Any, Dict, Iterable, Optional
from urllib.error import HTTPError

import jwt
from cryptography.fernet import Fernet, InvalidToken
from jwt import PyJWK, PyJWKClient, PyJWKClientError

from flask_cognito_lib.config import Config
from flask_cognito_lib.exceptions import CognitoError, TokenVerifyError


class TokenService:
    def __init__(self, cfg: Config):
        self.cfg = cfg
        self.jwk = PyJWKClient(self.cfg.jwk_endpoint, cache_keys=True)
        self.fernet = Fernet(self.get_encryption_key(self.cfg))

    @staticmethod
    def get_encryption_key(cfg) -> bytes:
        """Get the encryption key from the Flask `SECRET_KEY` for the Fernet cipher

        Returns
        -------
        bytes
            The encryption key
        """
        return urlsafe_b64encode(sha256(cfg.secret_key).digest())

    def get_public_key(self, token: str) -> PyJWK:
        """Find the public key ID for a given JWT

        Parameters
        ----------
        token : str
            The access token in JWT format. Must have `kid` in headers.

        Returns
        -------
        PyJWK
            A PyJWK instance that contains the public information

        Raises
        ------
        CognitoError
            If the PyJWTClient raises or the request to the user pool JWK
            endpoint fails
        """
        try:
            return self.jwk.get_signing_key_from_jwt(token)
        except (PyJWKClientError, HTTPError) as err:
            raise CognitoError("Error getting public keys from Cognito") from err

    def _jwt_validate(
        self,
        token: str,
        options: Dict[str, bool],
        leeway: float = 0,
        required: Optional[Iterable[str]] = None,
    ) -> Dict[str, Any]:
        """Validate the contents and claims of a JSON Web Token (JWT)

        Parameters
        ----------
        token : str
            Token in JWT format
        leeway : float, optional
            Leeway in seconds if validating expiry, by default 0
        options : Dict[str, bool]
            extended decoding and validation options, by default None
            See ``pyjwt.jwt.decode`` for options.
        required : Iterable[str], optional
            List of claims that must be present.
            Will set "aud", "iss", "exp", "iat" by default.

        Returns
        -------
        Dict[str, Any]
            Verified claims from the JWT

        Raises
        ------
        TokenVerifyError
            If claims or signature are invalid
        """
        try:
            claims = jwt.decode(
                jwt=token,
                key=self.get_public_key(token).key,
                algorithms=["RS256"],
                audience=self.cfg.user_pool_client_id,
                issuer=self.cfg.issuer,
                leeway=leeway,
                options=options,
                required=required or ["aud", "iss", "exp", "iat"],
            )

        except jwt.PyJWTError as err:
            raise TokenVerifyError("Token is not valid") from err

        return claims

    def verify_access_token(
        self,
        token: str,
        leeway: float = 0,
    ) -> Dict[str, Any]:
        """Verify the claims & signature of a JWT access token from Cognito

        This will check the audience, issuer, expiry and validate the signature
        of the JWT matches the public keys from the user pool

        Parameters
        ----------
        access_token : str
            The encoded JWT from Cognito
        leeway : float
            A time margin in seconds for the expiration check

        Returns
        -------
        Dict[str, Any]
            The verified claims from the encoded JWT

        Raises
        ------
        TokenVerifyError
            If not token is passed, or any checks fail
        """
        if not token:
            raise TokenVerifyError("No token provided")

        # Verify the contents and signature of the JWT
        claims = self._jwt_validate(
            token=token,
            leeway=leeway,
            options={
                "verify_signature": True,
                "verify_aud": False,  # JWT from Cognito will set client_id
                "verify_iss": True,
                "verify_exp": True,
                "verify_iat": True,
                "verify_nbf": False,  # Not issued
            },
        )

        # Cognito does not set an audience, but should populate client_id
        if claims["client_id"] != self.cfg.user_pool_client_id:
            raise TokenVerifyError("Token was not issued for this client id")

        return claims

    def verify_id_token(
        self,
        token: str,
        leeway: float = 0,
        nonce: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Verify the claims & signature of an id token in JWT format from Cognito

        This will check the audience, issuer, expiry and validate the signature
        of the JWT matches the public keys from the user pool

        Parameters
        ----------
        token : str
            The encoded JWT
        leeway : flaot
            A time margin in seconds for the expiration check
        nonce : Optional[str]
            An optional nonce value to validate to prevent replay attacks

        Returns
        -------
        Dict[str, Any]
            The OIDC claims from the encoded JWT

        Raises
        ------
        TokenVerifyError
            If not token is passed, or any checks fail
        """
        if not token:
            raise TokenVerifyError("No token provided")

        # Verify the contents and signature of the JWT
        claims = self._jwt_validate(
            token=token,
            leeway=leeway,
            options={
                "verify_signature": True,
                "verify_aud": True,  # ID token does contain correct audience
                "verify_iss": True,
                "verify_exp": True,
                "verify_iat": True,
                "verify_nbf": False,  # Not issued
            },
        )

        # Check nonce value to prevent replay attacks
        if nonce and claims["nonce"] != nonce:
            raise TokenVerifyError("Token nonce check failed")

        return claims

    def encrypt_token(self, token: str) -> str:
        """Symmetrically encrypt a token using Fernet with the Flask `SECRET_KEY`

        Parameters
        ----------
        token : str
            The token to encrypt

        Returns
        -------
        str
            The encrypted token
        """
        return self.fernet.encrypt(token.encode()).decode()

    def decrypt_token(self, token: str) -> str:
        """Decrypt a Fernet encrypted token using the Flask `SECRET_KEY`

        Parameters
        ----------
        token : str
            The token to decrypt

        Returns
        -------
        str
            The decrypted token

        Raises
        ------
        CognitoError
            If the token cannot be decrypted
        """
        try:
            return self.fernet.decrypt(token.encode()).decode()
        except InvalidToken as err:
            raise CognitoError("Error decrypting token") from err



from typing import Any, Callable, Dict, Optional

from flask import Flask, g

from flask_cognito_lib.config import Config
from flask_cognito_lib.exceptions import CognitoError
from flask_cognito_lib.services import cognito_service_factory, token_service_factory
from flask_cognito_lib.services.cognito_svc import CognitoService
from flask_cognito_lib.services.token_svc import TokenService
from flask_cognito_lib.utils import CognitoTokenResponse


class CognitoAuth:
    def __init__(
        self,
        app: Optional[Flask] = None,
        _token_service_factory: Callable = token_service_factory,
        _cognito_service_factory: Callable = cognito_service_factory,
        cfg: Optional[Config] = None,
    ):
        """Instantiate the CognitoAuth manager

        Parameters
        ----------
        app : Optional[Flask], optional
            An optional instance of a Flask application. If doing lazy init
            use the `init_app` method instead
        cfg : Optional[Config], optional
            Configuration object to use. If not provided, a default Config is used.
        """
        self.token_service_factory = _token_service_factory
        self.cognito_service_factory = _cognito_service_factory
        if app is not None:
            self.init_app(app=app, cfg=cfg)

    def init_app(self, app: Flask, cfg: Optional[Config] = None):
        """Register the extension with a Flask application

        Parameters
        ----------
        app : Flask
            Flask application
        cfg : Optional[Config], optional
            Configuration object to use. If not provided, a default Config is used.
        """
        if cfg:
            self.cfg = cfg
        else:
            self.cfg = Config()
        app.extensions[self.cfg.APP_EXTENSION_KEY] = self

    @property
    def token_service(self) -> TokenService:
        """Instantiate an instance of the TokenService within the app context

        Returns
        -------
        TokenService
            An instance of TokenService
        """
        if not hasattr(g, self.cfg.CONTEXT_KEY_TOKEN_SERVICE):
            token_service = self.token_service_factory(cfg=self.cfg)
            setattr(g, self.cfg.CONTEXT_KEY_TOKEN_SERVICE, token_service)
        return getattr(g, self.cfg.CONTEXT_KEY_TOKEN_SERVICE)

    @property
    def cognito_service(self) -> CognitoService:
        """Instantiate an instance of the CognitoService within the app context

        Returns
        -------
        CognitoService
            An instance of CognitoService
        """
        if not hasattr(g, self.cfg.CONTEXT_KEY_COGNITO_SERVICE):
            cognito_service = self.cognito_service_factory(cfg=self.cfg)
            setattr(g, self.cfg.CONTEXT_KEY_COGNITO_SERVICE, cognito_service)
        return getattr(g, self.cfg.CONTEXT_KEY_COGNITO_SERVICE)

    def get_tokens(
        self,
        request_args: Dict[str, str],
        expected_state: str,
        code_verifier: str,
    ) -> CognitoTokenResponse:
        """Exchange a short lived authorisation code for with Cognito for tokens

        Parameters
        ----------
        request_args : Dict[str, str]
            Request arguments returned from Cognito in the front channel
            i.e. URL query parameters. Should contain "code" and "state"
        expected_state : str
            The state value that was passed to Cognito when redirecting to the
            Cognito hosted UI. It should be returned unchanged in the
            ``request_args``
        code_verifier : str
            The plaintext code verification secret used as the code challenge
            when logging in

        Returns
        -------
        CognitoTokenResponse
            A dataclass that holds the token response from Cognito

        Raises
        ------
        CognitoError
            If access code or state or not in ``request_args``
            If the state value does not match the expected value
            If the request to the TOKEN endpoint fails
            If the TOKEN endpoint returns an error code
        """
        try:
            code = request_args["code"]
            state = request_args["state"]
        except KeyError as err:
            raise CognitoError(
                "Access code and/or state not returned from Cognito"
            ) from err

        if state != expected_state:
            raise CognitoError("State for CSRF is not correct")

        return self.cognito_service.exchange_code_for_token(
            code=code,
            code_verifier=code_verifier,
        )

    def exchange_refresh_token(
        self,
        refresh_token: str,
    ) -> CognitoTokenResponse:
        """Exchange a refresh token for a new set of tokens

        Parameters:
        -----------
        refresh_token : str
            The refresh token to exchange for a new set of tokens

        Returns:
        --------
        CognitoTokenResponse
            A dataclass that holds the token response from Cognito

        Raises:
        -------
        CognitoError
            If the request to the TOKEN endpoint fails
            If the TOKEN endpoint returns an error code
        """
        return self.cognito_service.exhange_refresh_token(
            refresh_token=refresh_token,
        )

    def revoke_refresh_token(
        self,
        refresh_token: str,
    ) -> None:
        """Revoke a refresh token

        Parameters:
        -----------
        refresh_token : str
            The refresh token to revoke

        Raises:
        -------
        CognitoError
            If the token isn't present in the request or
                if the feature is disabled for the app client
            If the token isn't a refresh token
            If the client credentials aren't valid
        """
        self.cognito_service.revoke_refresh_token(
            refresh_token=refresh_token,
        )

    def verify_access_token(self, token: str, leeway: float) -> Dict[str, Any]:
        """Verify the claims & signature of an access token in JWT format from Cognito

        This will check the audience, issuer, expiry and validate the signature
        of the JWT matches the public keys from the user pool

        Parameters
        ----------
        access_token : str
            The encoded JWT
        leeway : float
            A time margin in seconds for the expiration check

        Returns
        -------
        Dict[str, Any]
            The verified claims from the encoded JWT

        Raises
        ------
        TokenVerifyError
            If not token is passed, or any checks fail
        """
        return self.token_service.verify_access_token(token=token, leeway=leeway)

    def verify_id_token(
        self,
        token: str,
        leeway: float,
        nonce: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Verify the claims & signature of an id token in JWT format from Cognito

        This will check the audience, issuer, expiry and validate the signature
        of the JWT matches the public keys from the user pool

        Parameters
        ----------
        token : str
            The encoded JWT
        leeway : flaot
            A time margin in seconds for the expiration check
        nonce : Optional[str]
            An optional nonce value to validate to prevent replay attacks

        Returns
        -------
        Dict[str, Any]
            The OIDC claims from the encoded JWT

        Raises
        ------
        TokenVerifyError
            If not token is passed, or any checks fail
        """
        return self.token_service.verify_id_token(
            token=token,
            leeway=leeway,
            nonce=nonce,
        )



from functools import wraps
from typing import Iterable, Optional, Union

from flask import Response
from flask import current_app as app
from flask import redirect, request, session
from werkzeug.local import LocalProxy

from flask_cognito_lib.config import Config
from flask_cognito_lib.exceptions import (
    AuthorisationRequiredError,
    CognitoError,
    CognitoGroupRequiredError,
    TokenVerifyError,
)
from flask_cognito_lib.plugin import CognitoAuth
from flask_cognito_lib.utils import (
    CognitoTokenResponse,
    generate_code_challenge,
    generate_code_verifier,
    secure_random,
)

cognito_auth: CognitoAuth = LocalProxy(
    lambda: app.extensions[Config.APP_EXTENSION_KEY]
)  # type: ignore


def remove_from_session(keys: Iterable[str]):
    """Remove an entry from the session"""
    with app.app_context():
        for key in keys:
            if key in session:
                session.pop(key)


def validate_and_store_tokens(
    tokens: CognitoTokenResponse,
    nonce: Optional[str] = None,
) -> None:
    """Validate and store the access token and ID token (if present) in the session"""
    # validate the JWT and get the claims
    claims = cognito_auth.verify_access_token(
        token=tokens.access_token,
        leeway=cognito_auth.cfg.cognito_expiration_leeway,
    )
    session.update({"claims": claims})

    # Grab the user info from the user endpoint and store in the session
    if tokens.id_token is not None:
        user_info = cognito_auth.verify_id_token(
            token=tokens.id_token,
            nonce=nonce,
            leeway=cognito_auth.cfg.cognito_expiration_leeway,
        )
        session.update({"user_info": user_info})


def store_token_in_cookie(
    resp: Response,
    token: Union[str, None],
    cookie_name: str,
    max_age: int,
    encrypt: bool = False,
) -> None:
    """Store the token in an HTTP only secure cookie with the given name and max age

    Optionally symmetrically encrypt a token using Fernet with the Flask `SECRET_KEY`"""
    if encrypt:
        # Encrypt the token
        token = cognito_auth.token_service.encrypt_token(token)

    resp.set_cookie(
        key=cookie_name,
        value=token,
        max_age=max_age,
        httponly=True,
        secure=True,
        samesite=cognito_auth.cfg.cookie_samesite,
        domain=cognito_auth.cfg.cookie_domain,
    )


def get_token_from_cookie(cookie_name: str) -> Union[str, None]:
    """Get the token from the cookie"""
    token = request.cookies.get(cookie_name)

    if (
        token
        and cognito_auth.cfg.COOKIE_NAME_REFRESH == cookie_name
        and cognito_auth.cfg.refresh_cookie_encrypted
    ):
        # Decrypt the refresh token
        return cognito_auth.token_service.decrypt_token(token)

    return token


def cognito_login(fn):
    """A decorator that redirects to the Cognito hosted UI"""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        with app.app_context():
            # store parameters in the session that are passed to Cognito
            # and required for JWT verification
            code_verifier = generate_code_verifier()
            cognito_session = {
                "code_verifier": code_verifier,
                "code_challenge": generate_code_challenge(code_verifier),
                "nonce": secure_random(),
            }
            session.update(cognito_session)

            # Add suport for custom state values which are appended to a secure
            # random value for additional CRSF protection
            state = secure_random()
            custom_state = session.get("state")
            if custom_state:
                state += f"__{custom_state}"

            session.update({"state": state})

            login_url = cognito_auth.cognito_service.get_sign_in_url(
                code_challenge=session["code_challenge"],
                state=session["state"],
                nonce=session["nonce"],
                scopes=cognito_auth.cfg.cognito_scopes,
            )

        return redirect(login_url)

    return wrapper


def cognito_login_callback(fn):
    """
    A decorator to wrap the redirect after a user has logged in with Cognito.
    Stores the Cognito JWT in a http only cookie.
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        with app.app_context():
            # Get the access token return after auth flow with Cognito
            code_verifier = session["code_verifier"]
            state = session["state"]
            nonce = session["nonce"]

            # exchange the code for an access token
            # also confirms the returned state is correct
            tokens = cognito_auth.get_tokens(
                request_args=request.args,
                expected_state=state,
                code_verifier=code_verifier,
            )

            # Store the tokens in the session
            validate_and_store_tokens(tokens=tokens, nonce=nonce)

            # Remove one-time use variables now we have completed the auth flow
            remove_from_session(("code_challenge", "code_verifier", "nonce"))

            # split out the random part of the state value (in case the user
            # specified their own custom state value)
            state = session.get("state").split("__")[-1]
            session.update({"state": state})

            # return and set the JWT as a http only cookie
            resp = fn(*args, **kwargs)

            # Store the access token in a HTTP only secure cookie
            store_token_in_cookie(
                resp=resp,
                token=tokens.access_token,
                cookie_name=cognito_auth.cfg.COOKIE_NAME,
                max_age=cognito_auth.cfg.max_cookie_age_seconds,
            )

            # Grab the refresh token and store in a HTTP only secure cookie
            if cognito_auth.cfg.refresh_flow_enabled and tokens.refresh_token:
                store_token_in_cookie(
                    resp=resp,
                    token=tokens.refresh_token,
                    cookie_name=cognito_auth.cfg.COOKIE_NAME_REFRESH,
                    max_age=cognito_auth.cfg.max_refresh_cookie_age_seconds,
                    encrypt=cognito_auth.cfg.refresh_cookie_encrypted,
                )

        return resp

    return wrapper


def cognito_refresh_callback(fn):
    """A decorator that handles token refresh with Cognito"""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        with app.app_context():
            if not cognito_auth.cfg.refresh_flow_enabled:
                raise CognitoError("Refresh flow is not enabled")

            refresh_token = get_token_from_cookie(cognito_auth.cfg.COOKIE_NAME_REFRESH)

            if not refresh_token:
                raise CognitoError("No refresh token provided")

            # Exchange refresh token for the new access token.
            tokens = cognito_auth.exchange_refresh_token(
                refresh_token=refresh_token,
            )

            # Store the tokens in the session
            validate_and_store_tokens(tokens=tokens)

            # Return and set the JWT as a http only cookie
            resp = fn(*args, **kwargs)

            # Store the access token in a HTTP only secure cookie
            store_token_in_cookie(
                resp=resp,
                token=tokens.access_token,
                cookie_name=cognito_auth.cfg.COOKIE_NAME,
                max_age=cognito_auth.cfg.max_cookie_age_seconds,
            )

        return resp

    return wrapper


def cognito_logout(fn):
    """A decorator that handles logging out with Cognito"""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        with app.app_context():
            # logout at cognito and remove the cookies
            resp = redirect(cognito_auth.cfg.logout_endpoint)
            resp.delete_cookie(
                key=cognito_auth.cfg.COOKIE_NAME, domain=cognito_auth.cfg.cookie_domain
            )

            # Revoke the refresh token if it exists
            if refresh_token := get_token_from_cookie(
                cognito_auth.cfg.COOKIE_NAME_REFRESH
            ):
                cognito_auth.revoke_refresh_token(refresh_token)
                resp.delete_cookie(
                    key=cognito_auth.cfg.COOKIE_NAME_REFRESH,
                    domain=cognito_auth.cfg.cookie_domain,
                )

        # Cognito will redirect to the sign-out URL (if set) or else use
        # the callback URL
        return resp

    return wrapper


def auth_required(groups: Optional[Iterable[str]] = None, any_group: bool = False):
    """A decorator to protect a route with AWS Cognito"""

    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            with app.app_context():
                # return early if the extension is disabled
                if cognito_auth.cfg.disabled:
                    return fn(*args, **kwargs)

                # Try and validate the access token stored in the cookie
                try:
                    access_token = request.cookies.get(cognito_auth.cfg.COOKIE_NAME)
                    claims = cognito_auth.verify_access_token(
                        token=access_token,
                        leeway=cognito_auth.cfg.cognito_expiration_leeway,
                    )
                    valid = True

                    # Check for required group membership
                    if groups:
                        if any_group:
                            valid = any(g in claims["cognito:groups"] for g in groups)
                        else:
                            valid = all(g in claims["cognito:groups"] for g in groups)

                        if not valid:
                            raise CognitoGroupRequiredError

                except (TokenVerifyError, KeyError):
                    valid = False

                if valid:
                    return fn(*args, **kwargs)

                raise AuthorisationRequiredError

        return decorator

    return wrapper



import re
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from hashlib import sha256
from os import urandom
from typing import Optional


def secure_random(n_bytes: int = 32) -> str:
    """Generate a secure URL-safe random string"""
    return urlsafe_b64encode(urandom(n_bytes)).decode("utf-8")


def generate_code_verifier(n_bytes: int = 32) -> str:
    """Create a code verification secret"""
    code_verifier = secure_random(n_bytes=n_bytes)
    code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
    return code_verifier


def generate_code_challenge(code_verifier: str) -> str:
    """Create a code challenge (SHA256) from a code verifier"""
    code_challenge = sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = urlsafe_b64encode(code_challenge).decode("utf-8")
    return code_challenge.replace("=", "")


@dataclass
class CognitoTokenResponse:
    access_token: Optional[str] = None
    token_type: Optional[str] = None
    expires_in: Optional[int] = None
    refresh_token: Optional[str] = None
    id_token: Optional[str] = None
    error: Optional[str] = None



from typing import Dict, List, Optional
from urllib.parse import quote

import requests
from requests import JSONDecodeError, Response

from flask_cognito_lib.config import Config
from flask_cognito_lib.exceptions import CognitoError
from flask_cognito_lib.utils import CognitoTokenResponse


class CognitoService:
    def __init__(
        self,
        cfg: Config,
    ):
        self.cfg = cfg

    def get_sign_in_url(
        self,
        code_challenge: str,
        state: str,
        nonce: str,
        scopes: Optional[List[str]] = None,
    ) -> str:
        """Generate a sign URL against the AUTHORIZE endpoint

        Parameters
        ----------
        code_challenge : str
            A SHA256 hash of the code verifier used for this request.
            Note only S256 is support by AWS Cognito.
        state : str
            A random state string used for to prevent cross site request forgery
        nonce : str
            A random state string used for to prevent replay attacks
        scopes : Optional[List[str]]
            An optional list of system-reserved scopes or custom scopes that
            are associated with a client that can be requested.
            If the client doesn't request any scopes, the authentication server
            uses all scopes that are associated with the client.

        Returns
        -------
        str
            A front channel login URL for the AWS Cognito AUTHORIZE endpoint
        """
        quoted_redirect_url = quote(self.cfg.redirect_url)

        full_url = (
            f"{self.cfg.authorize_endpoint}"
            f"?response_type=code"
            f"&client_id={self.cfg.user_pool_client_id}"
            f"&redirect_uri={quoted_redirect_url}"
            f"&state={state}"
            f"&nonce={nonce}"
            f"&code_challenge={code_challenge}"
            "&code_challenge_method=S256"
        )

        if scopes is not None:
            full_url += f"&scope={'+'.join(scopes)}"

        return full_url

    def exchange_code_for_token(
        self,
        code: str,
        code_verifier: str,
    ) -> CognitoTokenResponse:
        """Exchange a short lived authorisation code for an access token

        Parameters
        ----------
        code : str
            The authorisation code after the user has logged in at the Cognito UI
        code_verifier : str
            The plaintext code verification secret used as the code challenge
            when logging in

        Returns
        -------
        CognitoTokenResponse
            A dataclass that holds the token response from Cognito

        Raises
        ------
        CognitoError
            If the request to the endpoint fails
            If the endpoint returns an error code
        """
        data = {
            "grant_type": "authorization_code",
            "client_id": self.cfg.user_pool_client_id,
            "redirect_uri": self.cfg.redirect_url,
            "code": code,
            "code_verifier": code_verifier,
        }

        return self._request_token(data)

    def exhange_refresh_token(
        self,
        refresh_token: str,
    ) -> CognitoTokenResponse:
        """Exchange a refresh token for a new set of tokens

        Parameters:
        -----------
        refresh_token : str
            The refresh token to exchange for a new set of
            ID and access tokens

        Returns:
        --------
        CognitoTokenResponse
            A dataclass that holds the token response from Cognito

        Raises:
        -------
        CognitoError
            If the request to the endpoint fails
            If the endpoint returns an error code
        """
        data = {
            "grant_type": "refresh_token",
            "client_id": self.cfg.user_pool_client_id,
            "refresh_token": refresh_token,
        }

        return self._request_token(data)

    def revoke_refresh_token(
        self,
        refresh_token: str,
    ) -> None:
        """Revoke a refresh token

        Parameters:
        -----------
        refresh_token : str
            The refresh token to revoke

        Raises:
        -------
        CognitoError
            If the token isn't present in the request or
                if the feature is disabled for the app client
            If the token isn't a refresh token
            If the client credentials aren't valid
        """
        data = {
            "token": refresh_token,
        }

        self._request(url=self.cfg.revoke_endpoint, data=data)

    def _request_token(self, data: Dict[str, str]) -> CognitoTokenResponse:
        """Request a token from the Cognito token endpoint

        Parameters
        ----------
        data : Dict[str, str]
            The data to be sent as part of the request

        Returns
        -------
        CognitoTokenResponse
            A dataclass that holds the token response from Cognito

        Raises:
        -------
        CognitoError
            If the request to the endpoint fails
            If the endpoint returns an error code
        """
        response = self._request(url=self.cfg.token_endpoint, data=data)

        try:
            return CognitoTokenResponse(**response.json())
        except JSONDecodeError as e:
            raise CognitoError(str(e)) from e

    def _request(self, url: str, data: Dict[str, str]) -> Response:
        """Make a request to the Cognito endpoint

        Parameters
        ----------
        url : str
            The URL of the endpoint
        data : Dict[str, str]
            The data to be sent as part of the request

        Returns
        -------
        Response
            The response from the endpoint

        Raises:
        -------
        CognitoError
            If the request to the endpoint fails
            If the endpoint returns an error code
        """
        # The Authorization header must not be present when using a
        # Public Client, we assume this when the secret is blank.
        # (Blank secrets are not supported on Confidential Clients)
        if self.cfg.user_pool_client_secret:
            auth = (self.cfg.user_pool_client_id, self.cfg.user_pool_client_secret)
        else:
            auth = None

        try:
            response = requests.post(
                url=url,
                data=data,
                auth=auth,
            )

        except requests.exceptions.RequestException as e:
            raise CognitoError(str(e)) from e

        try:
            response_json = response.json()
            if "error" in response_json:
                error_message = f"CognitoError: {response_json['error']}"

                if "error_description" in response_json:
                    error_message += f" - {response_json['error_description']}"

                raise CognitoError(error_message)
        except JSONDecodeError:
            # Some responses from Cognito are not JSON,so we can ignore this here.
            pass

        return response



from werkzeug.exceptions import HTTPException


class FlaskCognitoError(Exception):
    pass


class ConfigurationError(FlaskCognitoError):
    pass


class TokenVerifyError(FlaskCognitoError):
    pass


class CognitoError(FlaskCognitoError):
    pass


class AuthorisationRequiredError(HTTPException):
    code = 403
    description = "Authorization is required to access this resource."


class CognitoGroupRequiredError(HTTPException):
    code = 403
    description = "Cognito group membership is required to access this resource."
